(function () {

  /* global jasmineRequire, JsDiff */
  'use strict';

  // background colors for console
  // use "\x1B[K" for RESET to clear the line backgorund color
  // because there might be some colored whitespace if terminal scrolls the view
  var COLORS = {
    RESET: '\x1B[49m\x1B[K',
    GREEN: '\x1B[42m',
    RED: '\x1B[41m'
  };

  function wrapActual(str) {
    return COLORS.GREEN + str + COLORS.RESET;
  }

  function wrapExpected(str) {
    return COLORS.RED + str + COLORS.RESET;
  }

  function createDiffMessage(specData) {

    // try to find out if we are modifying a correct message

    var patterns = {
      toEqual: /Expected (.*) to equal (.*)./,
      toHaveBeenCalledWith: /Expected spy .* to have been called with (.*) but actual calls were (.*)./,
      toThrow: /Expected function to throw (.*), but it threw (.*)./,
      toThrowError: /Expected function to throw (.*), but it threw (.*)./
    };

    var pattern = patterns[specData.matcherName];
    if (!pattern) {
      // this spec matcher is not the one we want to modify
      return false;
    }

    // extract expected and actual objects (their stringified versions)
    // .exec returns null if there is no match
    var match = pattern.exec(specData.message) || [];
    var expected = match[1], actual = match[2];

    if (! (expected && actual)) {
      // seems that message is not the one we are looking for
      return false;
    }

    // toEqual displays objects in reverse order compared to other matchers
    var toEqual = specData.matcherName === 'toEqual';
    if (toEqual) {
      expected = match[2]; actual = match[1];
    }

    // diff by words, because it is more readable than diffChars
    var diff = JsDiff.diffWords(expected, actual);

    var actualDiff = '', expectedDiff = '';

    diff.forEach(function (part) {

      if (part.added) {

        actualDiff += wrapActual(part.value);

      } else if (part.removed) {

        expectedDiff += wrapExpected(part.value);

      } else {

        // add unmodified part to both outputs
        expectedDiff += part.value;
        actualDiff += part.value;

      }

    });

    // compose new message
    // replace original stringified objects with respective diffs

    var message = specData.message;

    message = message.replace(expected, expectedDiff);
    message = message.replace(actual, actualDiff);

    return message;
  }

  /*
   * And this is where black magic happens
   * Grabbing wrapper from Jasmine core for a Spec constructor, then
   * unwrapping it and replacing a method responsible for expectation results
   */

  var originalSpec = jasmineRequire.Spec;

  // rewrite wrapper
  jasmineRequire.Spec = function ($j) {

    // unwrap to get a constructor
    var Spec = originalSpec($j);

    var originalAddExpectationResult = Spec.prototype.addExpectationResult;

    Spec.prototype.addExpectationResult = function (passed, data) {

      if (!passed) {
        // message will be modified by reference
        // "createDiffMessage" may return falsy values, means that it failed
        // then use original message
        data.message = createDiffMessage(data) || data.message;
      }

      // call as it would originally be called, but with modified message
      originalAddExpectationResult.apply(this, arguments);
    };

    // return modified constructor for a wrapper
    return Spec;
  };

}());